/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: ManagedDatabaseImpl.java,v 1.6 2003/12/11 11:48:54 vgritsenko Exp $
 */

package org.apache.xindice.client.xmldb.managed;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.client.xmldb.CommonConfigurable;
import org.apache.xindice.client.xmldb.embed.CollectionImpl;
import org.apache.xindice.core.Database;
import org.apache.xindice.util.XindiceException;

import org.xmldb.api.base.Collection;
import org.xmldb.api.base.ErrorCodes;
import org.xmldb.api.base.XMLDBException;

import java.io.FileNotFoundException;

/**
 * The xindice-managed is a xindice client driver that requires third party code
 * to create the database within the constraints of the client classloader. Typically
 * it is used in embedded environments where database creation is managed by the third
 * party code.
 *
 * The driver will look for a database with the name passed in the constructor. If the no
 * argument constructor is used then the name defaults to 'db' though this can be changed
 * by setting {@link #DATABASE_NAME_PROPERTY} system property.
 *
 * This driver will throw an exception if the database instance hasn't been registered.
 *
 * @author <a href="mailto:kevin@rocketred.com.au">Kevin O'Neill</a>
 * @version CVS $Revision: 1.6 $, $Date: 2003/12/11 11:48:54 $
 */
public class ManagedDatabaseImpl extends CommonConfigurable implements org.xmldb.api.base.Database {

    /**
     * Logger used by this class
     */
    private static final Log log = LogFactory.getLog(ManagedDatabaseImpl.class);

    /**
     * Prefix used to denote XML:DB URI's that should use this driver
     */
    public static final String DRIVER_NAME = "xindice-managed";

    /**
     * Default database name in case none was specified via
     * {@link #DATABASE_NAME_PROPERTY} property
     */
    public static final String DATABASE_NAME = "db";

    /**
     * System property storing the name of the database to connect to
     */
    public static final String DATABASE_NAME_PROPERTY = "xindice.drivers.managed.db";

    /**
     * XML:DB conformance level of this driver
     */
    public static final String CONFORMANCE_LEVEL = "0";

    /**
     * Database instance
     */
    protected Database database;


    public ManagedDatabaseImpl() throws FileNotFoundException, XindiceException {
        this(System.getProperty(DATABASE_NAME_PROPERTY, DATABASE_NAME));
    }

    public ManagedDatabaseImpl(String databaseName) throws FileNotFoundException, XindiceException {
        this.database = Database.getDatabase(databaseName);

        if (null == database) {
            log.fatal("The database " + databaseName + " has not been created.");
            throw new XindiceException("The " + getName() + " driver requires that the database "
                                       + "instance be created and available in the JVM");
        }
    }

    /* (non-Javadoc)
    * @see org.xmldb.api.base.Database#getName()
    */
    public String getName() {
        return DRIVER_NAME;
    }

    /**
     * Retrieves a <code>Collection</code> instance based on the URI provided
     * in the <code>uri</code> parameter. The format of the URI is defined in the
     * documentation for DatabaseManager.getCollection().<p/>
     */
    public Collection getCollection(String uri, String username, String password) throws XMLDBException {
        /* TODO: introduce authentication some day */

        if (!acceptsURI(uri)) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL: " + uri);
        }

        /* Chop off driver prefix, and '://' */
        uri = uri.substring(getName().length() + 3);

        /* Extract host name & port, if present */
        int firstSlash = uri.indexOf('/');
        if (firstSlash == -1) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL (must have '/'): " + uri);
        }

        /* Extract collection name */
        String collPath = uri.substring(firstSlash);
        if (!collPath.startsWith("/")) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL (collection name must start with '/'): " + uri);
        }

        // find the database name. We just skip the first slash
        int colIndex = collPath.indexOf('/', 1);

        // We assume there's no collection specified
        String dbName = collPath.substring(1);
        String colName = "/";

        // if colIndex isn't -1 then we need to pick out the db and collection
        if (colIndex != -1) {
            dbName = collPath.substring(1, colIndex);

            // The rest of the name locates the collection
            colName = collPath.substring(colIndex);
            if (colName.equals("")) {
                colName = "/";
            }
        }

        // TODO: Is this correct behavior?
        if (!database.getName().equals(dbName)) {
            throw new XMLDBException(ErrorCodes.NO_SUCH_DATABASE,
                                     "Unknown database (must be '" + database.getName() + "'): " + uri);
        }

        try {
            return new CollectionImpl(database, colName);
        } catch (XMLDBException xmldbe) {
            if (xmldbe.errorCode == ErrorCodes.NO_SUCH_COLLECTION) {
                // per getCollection contract, return null if not found
                return null;
            } else {
                throw xmldbe;
            }
        }
    }

    public boolean acceptsURI(String uri) throws XMLDBException {
        return ((uri != null) && uri.startsWith(getName() + "://"));
    }

    public String getConformanceLevel() throws XMLDBException {
        return CONFORMANCE_LEVEL;
    }
}
