/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: MetaSystemCollection.java,v 1.11 2003/08/09 21:19:53 vladimir Exp $
 */

package org.apache.xindice.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.meta.MetaData;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.xml.dom.DOMParser;

import org.w3c.dom.Document;

import java.util.StringTokenizer;

/**
 * MetaSystemCollection represents the Meta System Collection.  Beyond
 * standard Collection operations, this class will provide facilities
 * for Meta data management.
 *
 * @version CVS $Revision: 1.11 $, $Date: 2003/08/09 21:19:53 $
 */
public final class MetaSystemCollection extends Collection {

    private static final Log log = LogFactory.getLog(MetaSystemCollection.class);

    public static final String METACOL = "meta";
    public static final String METAS = "Metas";
    public static final String COLLECTION_META_DATA = "_META_DATA_";

    private static String METACOL_DEFINITION
            = "<collection name=\"" + METACOL + "\">"
            // Meta System Collections
            + "   <collections>"
            //       Meta Collections
            + "      <collection name=\"" + METAS + "\" compressed=\"true\">"
            + "         <filer class=\"org.apache.xindice.core.filer.BTreeFiler\" />"
            + "         <collections>"
            //             Add Meta collections for System collections to avoid eternal loop during initialization
            +              SystemCollection.getDefinition()
            + "         </collections>"
            + "      </collection>"
            + "   </collections>"
            + "</collection>";

    private String dbCanonicalName;


    public MetaSystemCollection(Database db) {
        super(db);

        // Set the canonical names
        dbCanonicalName = db.getCanonicalName();
    }

    public void init() throws DBException {
        // Bootstrap the Meta Collection
        try {
            Document metaDoc = DOMParser.toDocument(METACOL_DEFINITION);
            Configuration metaCfg = new Configuration(metaDoc, false);
            setConfig(metaCfg);
        } catch (Exception e) {
            if (log.isFatalEnabled()) {
                log.fatal("FATAL ERROR: Generating System Collection '" + METACOL + "'", e);
            }
            System.exit(1);
        }
    }

    /**
     * Returns the corresponding Meta Collection for the given
     * collection, optionally creating the hierarchy
     *
     * @param collection the collection whose meta you want
     * @param create whether or not to create the meta if its missing
     * @return collection the meta collection for the requested collection
     */
    public Collection getMetaCollection(Collection collection, boolean create) throws DBException {
        String path = collection.getCanonicalName();

        // different database
        if (!path.startsWith(dbCanonicalName)) {
            return null;
        }

        // this is a meta collection
        if (path.startsWith(getCanonicalName())) {
            return null;
        }

        Collection current = getCollection(METAS);
        StringTokenizer st = new StringTokenizer(path.substring(dbCanonicalName.length()), "/");

        while (current != null && st.hasMoreTokens()) {
            String segment = st.nextToken().trim();
            if (segment.length() == 0) {
                continue;
            }

            Collection childcol = current.getCollection(segment);
            if (null == childcol) {
                if (!create) {
                    return null;
                }

                String cfgText
                        = "<collection name=\"" + segment + "\" compressed=\"true\">"
                        + "  <filer class=\"org.apache.xindice.core.filer.BTreeFiler\" />"
                        + "</collection>";

                try {
                    Document cfgDoc = DOMParser.toDocument(cfgText);
                    Configuration cfg = new Configuration(cfgDoc, false);
                    childcol = current.createCollection(segment, cfg);
                } catch (DBException de) {
                    throw de;
                } catch (Exception e) {
                    throw new DBException(FaultCodes.getFaultCode(e));
                }
            }
            current = childcol;
        }

        return current;
    }

    /**
     * dropCollectionMeta removes the MetaData for the specified
     * Collection.
     *
     * @param collection The Collection whose meta is required
     */
    public void dropCollectionMeta(Collection collection) throws DBException {
        Collection mcol = getMetaCollection(collection, false);
        if (null != mcol) {
            try {
                dropCollection(mcol);
            } catch (DBException e) {
                if (log.isWarnEnabled()) {
                    log.warn("ignored exception", e);
                }
            }
        }
    }

    /**
     * getCollectionMeta retrieves the MetaData for the specified
     * Collection.  Assumption is that the corresponding collection
     * must already exist.
     *
     * @param collection The Collection whose meta is required
     * @return The requested MetaData
     */
    public MetaData getCollectionMeta(Collection collection) throws DBException {
        Collection mcol = getMetaCollection(collection, true);

        if (null != mcol) {
            MetaData meta = (MetaData) mcol.getObject(COLLECTION_META_DATA);
            if (meta == null) {
                meta = new MetaData();
            }

            if (meta.getType() == MetaData.UNKNOWN) {
                meta.setType(MetaData.COLLECTION);
            }
            meta.setOwner(collection.getCanonicalName());
            meta.setDirty(false);
            return meta;
        }
        return null;
    }

    /**
     * setCollectionMeta save the Meta for the specified Collection.
     *
     * @param collection The Collection that owns the MetaData
     * @param meta The MetaData
     */
    public void setCollectionMeta(Collection collection, MetaData meta) throws DBException {
        if (null == meta) {
            return;
        }

        Collection mcol = getMetaCollection(collection, true);
        if (null != mcol) {
            mcol.setObject(COLLECTION_META_DATA, meta);
            meta.setDirty(false);
        }
    }

    /**
     * dropDocumentMeta removes the MetaData for the specified
     * Document.
     *
     * @param collection The Collection whose meta is required
     * @param id the Document id
     */
    public void dropDocumentMeta(Collection collection, String id) throws DBException {
        Collection mcol = getMetaCollection(collection, false);
        if (null != mcol) {
            try {
                mcol.remove(id);
            } catch (DBException e) {
                if (log.isWarnEnabled()) {
                    log.warn("ignored exception", e);
                }
            }
        }
    }

    /**
     * getDocumentMeta retrieves the MetaData for the specified
     * document.
     *
     * @param collection The Collection whose meta is required
     * @param id the ID
     * @return The requested MetaData
     */
    public MetaData getDocumentMeta(Collection collection, String id) throws DBException {
        Collection mcol = getMetaCollection(collection, true);
        if (null != mcol) {
            MetaData meta = (MetaData) mcol.getObject(id);

            if (meta == null) {
                meta = new MetaData();
            }

            if (meta.getType() == MetaData.UNKNOWN) {
                meta.setType(MetaData.DOCUMENT);
            }
            meta.setOwner(collection.getCanonicalName() + "/" + id);
            meta.setDirty(false);
            return meta;
        }
        return null;
    }

    /**
     * setDocumentMeta save the Meta for the specified Collection.
     *
     * @param collection The Collection that owns the MetaData
     * @param id
     * @param meta The MetaData
     */
    public void setDocumentMeta(Collection collection, String id, MetaData meta) throws DBException {
        if (null == meta) {
            return;
        }

        Collection mcol = getMetaCollection(collection, true);
        if (null != mcol) {
            mcol.setObject(id, meta);
            meta.setDirty(false);
        }
    }
}
