/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: InlineHeaderBuilder.java,v 1.5 2003/08/08 01:05:50 vgritsenko Exp $
 */

package org.apache.xindice.core.meta.inline;

import org.apache.xindice.core.data.Value;

/**
 * Construct a Value object containing the inline metadata header
 * and value data.
 *
 * The first header byte holds the number of total bytes (counting itself)
 * preceding the data.  The second header byte contains the version number
 * of the metadata.  Remaining header bytes are the version-specific metadata.
 *
 * <pre>
 * Version 0: Null
 *     no version-specific metdata, two byte total length
 *     byte 1: 2
 *     byte 2: 0
 *
 * Version 1: ResoureType
 *     resource type, three byte total length
 *     byte 1: 2
 *     byte 2: 1
 *     byte 3: content type (1 => xml, 2 => binary)
 * </pre>
 *
 * @version CVS $Revision: 1.5 $, $Date: 2003/08/08 01:05:50 $
 */
public class InlineHeaderBuilder {

    /**
     * There's no reason to ever create an instance of this class.
     */
    private InlineHeaderBuilder() {
    }

    /**
     * Create a Value object containing the inline metadata header and
     * value data.
     *
     * NOTE: For maximum performance, we would hand all the pieces to the Value
     * object,  which would talk to the BTree by stream,  and none of these
     * allocations and copies would be necessary.
     */
    public static Value createValue(int version, byte[] metadata, byte[] data, int pos, int len) {

        int headerLength = 2 + metadata.length;
        byte[] valueData = new byte[headerLength + len];

        valueData[0] = (byte) headerLength;
        valueData[1] = (byte) version;
        System.arraycopy(metadata, 0, valueData, 2, metadata.length);
        System.arraycopy(data, pos, valueData, 2 + metadata.length, len);
        return new Value(valueData);
    }
}
