/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: DatabaseTest.java,v 1.2 2003/12/24 14:10:04 vgritsenko Exp $
 */

package org.apache.xindice.core;

import org.apache.xindice.util.Configuration;
import org.apache.xindice.xml.dom.DOMParser;

import junit.framework.TestCase;

/**
 * Tests Xindice Core API (org.apache.xindice.core.Database,
 * org.apache.xindice.core.Collection) createCollection.
 *
 * @author <a href="mailto:vgritsenko@apache.org">Vadim Gritsenko</a>
 * @version CVS $Revision: 1.2 $, $Date: 2003/12/24 14:10:04 $
 */
public class DatabaseTest extends TestCase {

    public static final String DATABASE =
        "<root-collection dbroot=\"db/\" name=\"db\">" +
            "<queryengine>" +
                "<resolver autoindex=\"false\" class=\"org.apache.xindice.core.query.XPathQueryResolver\" />" +
                "<resolver class=\"org.apache.xindice.core.xupdate.XUpdateQueryResolver\" />" +
            "</queryengine>" +
        "</root-collection>";

    public static final String COLLECTIONA =
        "<collection compressed=\"true\" name=\"CollectionA\">" +
            "<filer class=\"org.apache.xindice.core.filer.BTreeFiler\" />" +
            "<indexes />" +
        "</collection>";

    public static final String COLLECTIONB =
        "<collection compressed=\"true\" name=\"CollectionB\">" +
            "<filer class=\"org.apache.xindice.core.filer.BTreeFiler\" />" +
            "<indexes />" +
        "</collection>";

    public static final String COLLECTIONC =
        "<collection compressed=\"true\" name=\"CollectionC\" inline-metadata=\"true\">" +
            "<filer class=\"org.apache.xindice.core.filer.BTreeFiler\" />" +
            "<indexes />" +
        "</collection>";

	private Database db;
    private Configuration configurationA;
    private Configuration configurationB;
    private Configuration configurationC;

    public DatabaseTest(String name) {
        super(name);
    }

    public void setUp() throws Exception {
        db = new Database();
        db.setConfig(new Configuration(DOMParser.toDocument(DATABASE)));
        configurationA = new Configuration(DOMParser.toDocument(COLLECTIONA));
        configurationB = new Configuration(DOMParser.toDocument(COLLECTIONB));
        configurationC = new Configuration(DOMParser.toDocument(COLLECTIONC));
    }

    public void testCreateCollectionNullName() throws Exception {
        try {
            db.createCollection(null, configurationA);
            fail("DBException expected");
        } catch (DBException e) {
            // Ok
        }
    }

    public void testCreateCollectionEmptyName() throws Exception {
        try {
            db.createCollection("", configurationA);
            fail("DBException expected");
        } catch (DBException e) {
            // Ok
        }
    }

    public void testCreateCollectionNullConfig() throws Exception {
        try {
            db.createCollection("name", null);
            fail("DBException expected");
        } catch (DBException e) {
            // Ok
        }
    }

    public void testCreateCollection() throws Exception {
        Collection collection = db.createCollection("CollectionA", configurationA);
        assertNotNull("Collection expected", collection);
        assertEquals("Collection name", "CollectionA", collection.getName());
    }

    public void testCreateDuplicateCollection() throws Exception {
        db.createCollection("CollectionA", configurationA);
        try {
            db.createCollection("CollectionA", configurationA);
            fail("DBException expected");
        } catch (DBException e) {
            // Ok
        }
    }

    public void testCreateCollectionWrongName() throws Exception {
        try {
            // Name does not match name in configuration
            db.createCollection("somepath", configurationA);
            fail("Exception expected");
        } catch (DBException e) {
            // Expected
        }
    }

    /**
     * Multithreaded getCollection test
     */
    public void testMTGetCollection() throws Exception {
        Collection a = db.createCollection("CollectionA", configurationA);
        Collection b = db.createCollection("CollectionB", configurationB);
        Collection c = db.createCollection("CollectionC", configurationC);

        final int THREADS = 20;
        final int ITERATIONS = 50;

        final boolean[] result = new boolean[]{ true };
        Thread[] threads = new Thread[THREADS];
        for (int i = 0; i < THREADS; i++) {
            threads[i] = new Thread() {
                public void run() {
                    for (int ii = 0; ii < ITERATIONS && result[0]; ii++) {
                        Collection c;
                        try {
                            c = db.getCollection("CollectionA");
                            if (!c.getName().equals("CollectionA")) {
                                System.out.println("Expected CollectionA, got " + c.getName());
                                result[0] = false;
                            }
                            c = db.getCollection("CollectionB");
                            if (!c.getName().equals("CollectionB")) {
                                System.out.println("Expected CollectionB, got " + c.getName());
                                result[0] = false;
                            }
                            c = db.getCollection("CollectionC");
                            if (!c.getName().equals("CollectionC")) {
                                System.out.println("Expected CollectionC, got " + c.getName());
                                result[0] = false;
                            }
                        } catch (Exception e) {
                            e.printStackTrace();
                            result[0] = false;
                            fail();
                        }
                    }
                }
            };
            threads[i].setName("DatabaseTest" + i);
        }

        // Start all the threads at once
        for (int i = 0; i < THREADS; i++) {
            threads[i].start();
        }
        Thread.sleep(1000);
        for (int i = 0; i < THREADS; i++) {
            threads[i].join();
        }

        db.dropCollection(a);
        db.dropCollection(b);
        db.dropCollection(c);

        assertTrue("Test Failed", result[0]);
    }

    public void tearDown() throws Exception {
        try {
            db.dropCollection(db.getCollection("CollectionA"));
        } catch (Exception e) {}
        try {
            db.dropCollection(db.getCollection("CollectionB"));
        } catch (Exception e) {}
        try {
            db.dropCollection(db.getCollection("CollectionC"));
        } catch (Exception e) {}
        db.close();
    }
}
