/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: XPathQueryResolverTest.java,v 1.6 2003/12/22 02:14:07 vgritsenko Exp $
 */

package org.apache.xindice.core.query;

import org.apache.xindice.core.Collection;
import org.apache.xindice.core.Database;
import org.apache.xindice.core.data.NodeSet;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.xml.dom.DOMParser;

import junit.framework.TestCase;
import org.w3c.dom.Element;
import org.w3c.dom.Text;

/**
 * This TestCase should test the correct Work of the XPathQueryResolver.
 *
 * During setUp it generates a database and adds one Collection.
 * The two Tests Simple and Complex query will be raun against this Collection.
 * The complex query string is taken from the Bug 14878 in Bugzilla
 *
 * @version CVS $Revision: 1.6 $, $Date: 2003/12/22 02:14:07 $
 * @author Alexander Sterff <sterff@in.tum.de>
 */
public class XPathQueryResolverTest extends TestCase {

    final String DATABASE =
        "<root-collection dbroot=\"db/\" name=\"db\">" +
            "<queryengine>" +
                "<resolver autoindex=\"false\" class=\"org.apache.xindice.core.query.XPathQueryResolver\" />" +
                "<resolver class=\"org.apache.xindice.core.xupdate.XUpdateQueryResolver\" />" +
            "</queryengine>" +
        "</root-collection>";

    final String COLLECTION =
        "<collection compressed=\"true\" name=\"testcol\">" +
            "<filer class=\"org.apache.xindice.core.filer.BTreeFiler\" />" +
            "<indexes />" +
        "</collection>";

    final String DOCUMENT =
        "<terrainmap>" +
            "<coordinates>" +
                "some text" +
                "<top-left>" +
                    "<latlong>" +
                        "some more text" +
                        "<latitude>80</latitude>" +
                        "<longitude>-100</longitude>" +
                    "</latlong>" +
                "</top-left>" +
                "<bottom-right>" +
                    "<latlong>" +
                        "a" +
                        "<latitude>-200</latitude>" +
                        "b" +
                        "<longitude>90</longitude>" +
                        "c" +
                    "</latlong>" +
                "</bottom-right>" +
            "</coordinates>" +
        "</terrainmap>";

    final String[] ELEMENT_QUERY = {
        "/terrainmap",
        "/terrainxmap",
        "/terrainmap/coordinates",
        "/terrainmap/coordinates/top-left",
        "/terrainmap/coordinates/*",
        "/terrainmap/coordinates/bottom-right/latlong",
        "/terrainmap/coordinates/bottom-right/latlong/*",
        "/terrainmap/coordinates/bottom-right/latlong/..",
    };

    final int[] ELEMENT_COUNT = {
        1,
        0,
        1,
        1,
        2,
        1,
        2,
        1,
    };

    final String[] ELEMENT_NAMES = {
        "terrainmap",
        null,
        "coordinates",
        "top-left",
        "bottom-right",
        "latlong",
        "longitude",
        "bottom-right",
    };

    final String[] NUMBER_QUERY = {
        "2 + 2",
        "count(/terrainmap)",
        "count(/terrainmap/coordinates) + count(//latlong)",
        "number(//latitude[1]/text())",
        "number(//longitude[ancestor::bottom-right]/text())",
    };

    final int[] NUMBER_VALUE = {
        4,
        1,
        3,
        80,
        90,
    };

    final String[] TEXT_QUERY = {
        "/terrainmap/coordinates/text()",
        "/terrainmap/coordinates//text()[contains(., 'more')]",
        "//text()",
        "//latlong/text()",
        "//latlong/text()[2]",
    };

    final int[] TEXT_COUNT = {
        1,
        1,
        9,
        4,
        1,
    };

    final String[] TEXT_VALUE = {
        "some text",
        "some more text",
        "c",
        "c",
        "b",
    };


    private static int tests;
	private static Database db;
	private static Collection col;

    private XPathQueryResolver resolver;


    public XPathQueryResolverTest(String name) {
        super(name);
        tests++;
    }

    public void setUp() throws Exception {
        if (db == null) {
            Database db = new Database();
            db.setConfig(new Configuration(DOMParser.toDocument(DATABASE)));
            col = db.createCollection("testcol", new Configuration(DOMParser.toDocument(COLLECTION)));
            col.insertDocument(DOMParser.toDocument(DOCUMENT));

            XPathQueryResolverTest.db = db;
        }

        resolver = new XPathQueryResolver();
    }

    public void testElementXPathQuery() throws Exception {
        for (int i = 0; i < ELEMENT_QUERY.length; i++) {
            NodeSet nodeSet = resolver.query(col, ELEMENT_QUERY[i], null, null);
            int results = 0;
            Element element = null;
            while(nodeSet.hasMoreNodes()) {
                element = (Element)nodeSet.getNextNode();
                assertNotNull("NodeSet.getNextNode must be not null", element);
                results++;
            }

            assertEquals("Results count", ELEMENT_COUNT[i], results);
            if (results > 0) {
                assertEquals("Result element name", ELEMENT_NAMES[i], element.getNodeName());
            }
        }
    }

    public void testNumberXPathQuery() throws Exception {
        for (int i = 0; i < NUMBER_QUERY.length; i++) {
            NodeSet nodeSet = resolver.query(col, NUMBER_QUERY[i], null, null);
            Double element = null;
            assertTrue("Expected one result", nodeSet.hasMoreNodes());
            element = (Double)nodeSet.getNextNode();
            assertNotNull("NodeSet.getNextNode must be not null", element);
            assertEquals("Result value", new Double(NUMBER_VALUE[i]), element);
        }
    }

    public void testTextXPathQuery() throws Exception {
        for (int i = 0; i < TEXT_QUERY.length; i++) {
            NodeSet nodeSet = resolver.query(col, TEXT_QUERY[i], null, null);
            int results = 0;
            Text element = null;
            while(nodeSet.hasMoreNodes()) {
                element = (Text)nodeSet.getNextNode();
                assertNotNull("NodeSet.getNextNode must be not null", element);
                results++;
            }

            assertEquals("Results count for " + TEXT_QUERY[i], TEXT_COUNT[i], results);
            if (results > 0) {
                assertEquals("Result value for " + TEXT_QUERY[i], TEXT_VALUE[i], element.getNodeValue());
            }
        }
    }

	public void testComplexXPathQuery() throws Exception {
		Object node = null;
		XPathQueryResolver queryResolv = new XPathQueryResolver();
		NodeSet nodeSet =
			queryResolv.query(
				col,
				"/terrainmap[coordinates/top-left/latlong/latitude[(number(text()) + 180) >= 0] and coordinates/top-left/latlong/longitude[(number(text()) + 90) <= 0] and coordinates/bottom-right/latlong/latitude[(number(text()) + 180) <= 0] and coordinates/bottom-right/latlong/longitude[(number(text()) + 90) >= 0]]",
				null,
				null);
		if (nodeSet.hasMoreNodes()) {
			node = nodeSet.getNextNode();
		}
		assertNotNull("Complex query didn't return a node", node);
	}

    public void tearDown() throws Exception {
        if (--tests == 0) {
            db.dropCollection(col);
        }
    }
}
